import datetime
import random

TEN_YEARS_AS_SECONDS = 10 * 365 * 24 * 3600


def sanitize_random_past_timestamp(value):
    num = random.randint(0, TEN_YEARS_AS_SECONDS * 1000)
    delta = datetime.timedelta(seconds=(num / 1000.0))
    dt = datetime.datetime.now() - delta
    return dt.isoformat()



from __future__ import unicode_literals

import mock
import pytest
import six

from database_sanitizer import __main__

main = __main__.main


@mock.patch.object(__main__, 'run')
def test_main_without_args(mocked_run, capsys):
    with pytest.raises(SystemExit) as excinfo:
        main(['SANI'])
    assert excinfo.value.code == 2

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err.splitlines() == [
        'usage: SANI [-h] [--config CONFIG] [--output OUTPUT] url',
        'SANI: error: the following arguments are required: url' if six.PY3
        else 'SANI: error: too few arguments',
    ]
    assert not mocked_run.called


@mock.patch.object(__main__, 'run')
def test_main_with_url(mocked_run, capsys):
    main(['SANI', 'some://url'])

    # Output should be empty
    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''

    # The run function should have been called with the URL
    (run_call_args, run_call_kwargs) = mocked_run.call_args
    assert run_call_args == ()
    assert set(run_call_kwargs.keys()) == {'config', 'output', 'url'}
    assert run_call_kwargs['config'] is None
    assert run_call_kwargs['url'] == 'some://url'


@pytest.mark.parametrize('optname', ['-c', '--config'])
@mock.patch.object(__main__, 'run')
@mock.patch.object(__main__, 'Configuration')
def test_main_with_config(mocked_conf, mocked_run, capsys, optname):
    main(['SANI', optname, 'config_file.yml', 'some://url'])

    # Output should be empty
    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''

    # Configuration should have been created with Configuration.from_file
    (fromfile_args, fromfile_kwargs) = mocked_conf.from_file.call_args
    assert fromfile_args == ('config_file.yml',)
    assert fromfile_kwargs == {}

    # The run function should have been called with the config and URL
    (run_call_args, run_call_kwargs) = mocked_run.call_args
    assert run_call_args == ()
    assert set(run_call_kwargs.keys()) == {'config', 'output', 'url'}
    assert run_call_kwargs['config'] == mocked_conf.from_file.return_value
    assert run_call_kwargs['url'] == 'some://url'


@pytest.mark.parametrize('optname', ['-o', '--output'])
@mock.patch.object(__main__, 'run')
@mock.patch.object(__main__, 'open')
def test_main_with_output(mocked_open, mocked_run, capsys, optname):
    main(['SANI', optname, 'output_file.sql', 'some://url'])

    # Output should be empty
    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''

    # Output file should have been opened
    (open_args, open_kwargs) = mocked_open.call_args
    assert open_args == ('output_file.sql', 'w')
    assert open_kwargs == {}

    # The run function should have been called with the output and URL
    (run_call_args, run_call_kwargs) = mocked_run.call_args
    assert run_call_args == ()
    assert set(run_call_kwargs.keys()) == {'config', 'output', 'url'}
    assert run_call_kwargs['config'] is None
    assert run_call_kwargs['output'] == mocked_open.return_value
    assert run_call_kwargs['url'] == 'some://url'



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import codecs
import io
import re
import subprocess

from ..utils.postgres import decode_copy_value, encode_copy_value
from ..config import PG_DUMP_DEFAULT_PARAMETERS

COPY_LINE_PATTERN = re.compile(
    r"^COPY \"(?P<schema>[^\"]*)\".\"(?P<table>[^\"]*)\" "
    r"\((?P<columns>.*)\) "
    r"FROM stdin;$"
)


def sanitize(url, config):
    """
    Obtains dump of an Postgres database by executing `pg_dump` command and
    sanitizes it's output.

    :param url: URL to the database which is going to be sanitized, parsed by
                Python's URL parser.
    :type url: six.moves.urllib.parse.ParseResult

    :param config: Optional sanitizer configuration to be used for sanitation
                   of the values stored in the database.
    :type config: database_sanitizer.config.Configuration|None
    """
    if url.scheme not in ("postgres", "postgresql", "postgis"):
        raise ValueError("Unsupported database type: '%s'" % (url.scheme,))

    extra_params = PG_DUMP_DEFAULT_PARAMETERS
    if config:
        extra_params = config.pg_dump_params

    process = subprocess.Popen(
        (
            "pg_dump",
            # Force output to be UTF-8 encoded.
            "--encoding=utf-8",
            # Quote all table and column names, just in case.
            "--quote-all-identifiers",
            # Luckily `pg_dump` supports DB URLs, so we can just pass it the
            # URL as argument to the command.
            "--dbname",
            url.geturl().replace('postgis://', 'postgresql://'),
         ) + tuple(extra_params),
        stdout=subprocess.PIPE,
    )

    sanitize_value_line = None
    current_table = None
    current_table_columns = None
    skip_table = False

    for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
        # Eat the trailing new line.
        line = line.rstrip("\n")

        # Are we currently in middle of `COPY` statement?
        if current_table:
            # Backslash following a dot marks end of an `COPY` statement.
            if line == "\\.":
                current_table = None
                current_table_columns = None
                if not skip_table:
                    yield "\\."
                skip_table = False
                continue

            if skip_table:
                continue

            if not sanitize_value_line:
                yield line
                continue

            yield sanitize_value_line(line)
            continue

        # Is the line beginning of `COPY` statement?
        copy_line_match = COPY_LINE_PATTERN.match(line)
        if not copy_line_match:
            yield line
            continue

        current_table = copy_line_match.group("table")
        current_table_columns = parse_column_names(copy_line_match.group("columns"))

        # Skip `COPY` statement if table rows are configured
        # to be skipped.
        if config and current_table in config.skip_rows_for_tables:
            skip_table = True
            continue

        sanitize_value_line = get_value_line_sanitizer(
            config, current_table, current_table_columns)

        yield line


def get_value_line_sanitizer(config, table, columns):
    if not config:
        return None

    def get_sanitizer(column):
        sanitizer = config.get_sanitizer_for(table, column)

        if not sanitizer:
            return _identity

        def decode_sanitize_encode(value):
            return encode_copy_value(sanitizer(decode_copy_value(value)))

        return decode_sanitize_encode

    sanitizers = [get_sanitizer(column) for column in columns]

    if all(x is _identity for x in sanitizers):
        return None

    def sanitize_line(line):
        values = line.split('\t')
        if len(values) != len(columns):
            raise ValueError("Mismatch between column names and values.")
        return '\t'.join(
            sanitizer(value)
            for (sanitizer, value) in zip(sanitizers, values))

    return sanitize_line


def _identity(x):
    return x


def parse_column_names(text):
    """
    Extracts column names from a string containing quoted and comma separated
    column names.

    :param text: Line extracted from `COPY` statement containing quoted and
                 comma separated column names.
    :type text: str

    :return: Tuple containing just the column names.
    :rtype: tuple[str]
    """
    return tuple(
        re.sub(r"^\"(.*)\"$", r"\1", column_name.strip())
        for column_name in text.split(",")
    )


def parse_values(text):
    """
    Parses line following `COPY` statement containing values for a single row
    in the table, in custom Postgres format.

    :param text: Line following `COPY` statement containing values.
    :type text: str

    :return: Column values extracted from the given line.
    :rtype: tuple[str|None]
    """
    return tuple(decode_copy_value(value) for value in text.split("\t"))



from database_sanitizer.sanitizers import constant


def test_sanitize_null():
    assert constant.sanitize_null(None) is None
    assert constant.sanitize_null('') is None
    assert constant.sanitize_null('whatever') is None
    assert constant.sanitize_null('test') is None


def test_sanitize_invalid_django_password():
    assert constant.sanitize_invalid_django_password(None) == '!'
    assert constant.sanitize_invalid_django_password('') == '!'
    assert constant.sanitize_invalid_django_password('whatever') == '!'
    assert constant.sanitize_invalid_django_password('test') == '!'


def test_sanitize_empty_json_dict():
    assert constant.sanitize_empty_json_dict(None) == '{}'
    assert constant.sanitize_empty_json_dict('') == '{}'
    assert constant.sanitize_empty_json_dict('whatever') == '{}'
    assert constant.sanitize_empty_json_dict('test') == '{}'


def test_sanitize_empty_json_list():
    assert constant.sanitize_empty_json_list(None) == '[]'
    assert constant.sanitize_empty_json_list('') == '[]'
    assert constant.sanitize_empty_json_list('whatever') == '[]'
    assert constant.sanitize_empty_json_list('test') == '[]'



import pytest

from faker import Faker

from db_anonymiser import faker as db_anonymiser_faker


# Force faker to use a seed when producing output so that we can assume a
# deterministic set of results
@pytest.fixture(autouse=True)
def seed_faker():
    Faker.seed(0)


def test_sanitize_name():
    assert db_anonymiser_faker.sanitize_name("Bob Benson") == "Dr Rhys Thomas"


def test_sanitize_first_name():
    assert db_anonymiser_faker.sanitize_first_name("Bob") == "Emma"


def test_sanitize_last_name():
    assert db_anonymiser_faker.sanitize_last_name("Benson") == "Watts"


def test_sanitize_email():
    assert (
        db_anonymiser_faker.sanitize_email("bob.benson@example.net")
        == "areed@example.org"
    )


def test_sanitize_company_name():
    assert db_anonymiser_faker.sanitize_company_name("Sterling Cooper") == "Reed-Burton"


def test_sanitize_phone_number():
    assert db_anonymiser_faker.sanitize_phone_number("011111111111") == "+448764759382"


def test_sanitize_address():
    assert (
        db_anonymiser_faker.sanitize_address("Madison Ave")
        == "Flat 4, Gibbons tunnel, Lesleystad, L8C 2EZ"
    )


def test_sanitize_website():
    assert (
        db_anonymiser_faker.sanitize_website("sterling.cooper")
        == "reed-burton.arnold-jones.com"
    )


def test_sanitize_text():
    assert (
        db_anonymiser_faker.sanitize_text("Not great, Bob.")
        == "Aliquam vitae laborum ullam rerum voluptas. Nesciunt tenetur magnam eligendi quidem nulla. Voluptates minus provident nobis corporis. Quas tempore placeat iusto. Explicabo et odit dignissimos."
    )
    assert db_anonymiser_faker.sanitize_text(None) == None


def test_sanitize_street_address():
    assert (
        db_anonymiser_faker.sanitize_street_address("Madison Ave")
        == "Studio 0\nArnold oval"
    )


def test_sanitize_city():
    assert db_anonymiser_faker.sanitize_city("New York") == "Reedchester"


def test_sanitize_postcode():
    assert db_anonymiser_faker.sanitize_postcode("DT11 7DY") == "TA60 8UR"


def test_sanitize_eori_number():
    assert (
        db_anonymiser_faker.sanitize_eori_number("GB111111111111") == "GB424533559245"
    )


def test_sanitize_ni_eori_number():
    assert (
        db_anonymiser_faker.sanitize_ni_eori_number("XI111111111111") == "XI424533559245"
    )

def test_sanitize_eu_eori_number():
    assert (
        db_anonymiser_faker.sanitize_eu_eori_number("FR11111") == "FR99346"
    )

def test_sanitize_sic_number():
    assert db_anonymiser_faker.sanitize_sic_number("11111") == "50494"


def test_sanitize_vat_number():
    assert db_anonymiser_faker.sanitize_vat_number("GB111111111") == "GB906691059"


def test_sanitize_registration_number():
    assert db_anonymiser_faker.sanitize_registration_number("11111111") == "51706749"


def test_sanitize_filename():
    assert db_anonymiser_faker.sanitize_filename("somefile.txt") == "molestiae.xlsx"


def test_sanitize_short_text():
    assert (
        db_anonymiser_faker.sanitize_short_text("some small text")
        == "Tempore placeat iusto aut. Et odit dignissimos mollitia ipsam maxime."
    )
    assert db_anonymiser_faker.sanitize_short_text(None) == None



import sys

from django.conf import settings
from django.core.management.base import BaseCommand

from . import _base_helpers as helpers
from . import _check_migration_fields_helpers as check_migration_fields_helpers


class Command(BaseCommand):
    help = "Check that new migration fields are in anonymiser config"

    def handle(self, *args, **options):
        writer = helpers.make_writers(self)
        config_path = settings.DB_ANONYMISER_CONFIG_LOCATION
        diff_content = check_migration_fields_helpers.get_diff_content()
        if not diff_content:
            return
        new_fields = check_migration_fields_helpers.extract_new_fields(diff_content)
        new_models = check_migration_fields_helpers.extract_new_models(diff_content)
        if not new_fields and not new_models:
            return
        # Handle AddField — check against config
        missing_fields = []
        if new_fields:
            missing_fields = check_migration_fields_helpers.check_fields_in_config(
                new_fields, config_path, writer
            )
            if missing_fields:
                writer.error("\n⚠️  New field(s) not found in anonymiser config:")
                for field in missing_fields:
                    writer.error(f"   - {field}")
                writer.warning("\nIf these fields contain sensitive data, add them to:")
                writer.warning(f"   {config_path}\n")
        # Handle CreateModel — always prompt
        if new_models:
            model_list = ", ".join(new_models)
            writer.warning(f"\n⚠️  New model(s) detected: {model_list}")
            writer.warning(
                "Please ensure any sensitive fields are added to the anonymiser config."
            )
        if missing_fields or new_models:
            response = input("Are you happy to proceed? (y/n): ").strip().lower()
            if response == "y":
                writer.success("\n✓ Proceeding with commit.\n")
                return
            elif response == "n":
                writer.error("\n❌ Commit blocked.\n")
                sys.exit(1)
            else:
                writer.error("\n❌ Invalid response. Commit blocked.\n")
                sys.exit(1)



# -*- coding: utf-8 -*-
"""
Contains utilities for working with Postgres `COPY` command, mainly encoding
and decoding values in the custom format used by Postgres.

Documentation about copy command and the text format used by it can be found
from:
https://www.postgresql.org/docs/9.2/static/sql-copy.html

For decoding we use a regular expression to find the escape sequences
and invoke `unescape_single_character` function for each occurence.
Allowed escape sequences are precalculated into `DECODE_MAP` to make the
lookups faster.

For encoding we use a string translation table `ENCODE_TRANSLATE_TABLE`,
which maps the "forbidden" characters to escape sequences.  This is used
with `str.translate`, which is very fast way to escape characters.
"""

from __future__ import unicode_literals

import itertools
import re

import six

#: Representation of NULL value in Postgres COPY statement.
POSTGRES_COPY_NULL_VALUE = "\\N"

ENCODE_MAP = {
    '\\': '\\\\',
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
    '\v': '\\v',
}

ENCODE_TRANSLATE_TABLE = [
    ENCODE_MAP.get(six.unichr(n), six.unichr(n))
    for n in range(256)
]

DECODE_REGEX = re.compile(r"""
\\                 # a backslash
(?:                # followed by one of these (in non-capturing parenthesis):
    [0-7]{1,3}         # 1, 2 or 3 octal digits
    |                  # or
    x[0-9a-fA-F]{1,2}  # 'x' followed by 1 or 2 hexadecimal digits
    |                  # or
    .                  # any character
    |                  # or
    \Z                 # end of string
)
""", re.VERBOSE)


def decode_copy_value(value):
    """
    Decodes value received as part of Postgres `COPY` command.

    :param value: Value to decode.
    :type value: str

    :return: Either None if the value is NULL string, or the given value where
             escape sequences have been decoded from.
    :rtype: str|None
    """
    # Test for null values first.
    if value == POSTGRES_COPY_NULL_VALUE:
        return None

    # If there is no backslash present, there's nothing to decode.
    #
    # This early return provides a little speed-up, because it's very
    # common to not have anything to decode and then simple search for
    # backslash is faster than the regex sub below.
    if '\\' not in value:
        return value

    return DECODE_REGEX.sub(unescape_single_character, value)


def unescape_single_character(match):
    """
    Unescape a single escape sequence found by regular expression.

    :param match: Regular expression match object
    :rtype: str
    :raises: ValueError if the escape sequence is invalid
    """
    try:
        return DECODE_MAP[match.group(0)]
    except KeyError:
        value = match.group(0)
        if value == '\\':
            raise ValueError("Unterminated escape sequence encountered")

        raise ValueError(
            "Unrecognized escape sequence encountered: {}".format(value))


def encode_copy_value(value):
    """
    Encodes given value into format suitable for Postgres `COPY` statement.

    :param value: Value to encode.
    :type value: str|None

    :return: Given value encoded into format that is suitable to be used in the
             `COPY` command.
    :rtype: str
    """
    if value is None:
        return POSTGRES_COPY_NULL_VALUE

    return value.translate(ENCODE_TRANSLATE_TABLE)


def _generate_decode_map():
    # Initialize the map by inverting the encode map
    decode_map = {
        encoded_char: char
        for (char, encoded_char) in ENCODE_MAP.items()
    }

    # Add entries for 1-3 octal digits and 1-2 hexadecimal digits
    digit_encode_params = [
        # (base, prefix, lengths, digit_chars)
        (8, '\\', [1, 2, 3], '01234567'),
        (16, '\\x', [1, 2], '0123456789abcdefABCDEF')
    ]
    for (base, prefix, lengths, digit_chars) in digit_encode_params:
        for length in lengths:
            for digits in itertools.product(digit_chars, repeat=length):
                digit_string = ''.join(digits)
                value = int(digit_string, base=base)
                char = six.unichr(value)
                decode_map[prefix + digit_string] = char

    return decode_map


DECODE_MAP = _generate_decode_map()



import uuid

from database_sanitizer.session import hash_text

NIL_UUID = '00000000-0000-0000-0000-000000000000'
NIL_UUID_WITHOUT_DASHES = NIL_UUID.replace('-', '')


def sanitize_uuid4(value):
    if not value:
        return value
    if value.replace('-', '') == NIL_UUID_WITHOUT_DASHES:
        return NIL_UUID
    return str(uuid.UUID(hash_text(value)[:32], version=4))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from ..utils.postgres import (
    DECODE_MAP,
    POSTGRES_COPY_NULL_VALUE,
    decode_copy_value,
    encode_copy_value,
)


@pytest.mark.parametrize(
    "input_value,expected_value",
    (
        ("", ""),
        (POSTGRES_COPY_NULL_VALUE, None),
        ("Test", "Test"),
        ("\\\\", "\\"),
        ("\\b", "\b"),
        ("\\f", "\f"),
        ("\\n", "\n"),
        ("\\r", "\r"),
        ("\\t", "\t"),
        ("\\v", "\v"),
        ("\\xff", "\xff"),
        ("\\123", "\123"),
        ("Test\\r\\nTest", "Test\r\nTest"),
    )
)
def test_decode_copy_value(input_value, expected_value):
    assert decode_copy_value(input_value) == expected_value


@pytest.mark.parametrize(
    "input_value,expected_value",
    (
        ("", ""),
        (None, POSTGRES_COPY_NULL_VALUE),
        ("Test", "Test"),
        ("\\", "\\\\"),
        ("\b", "\\b"),
        ("\f", "\\f"),
        ("\n", "\\n"),
        ("\r", "\\r"),
        ("\t", "\\t"),
        ("\v", "\\v"),
        ("\xff", "\xff"),
        ("\123", "\123"),
        ("Test\r\nTest", "Test\\r\\nTest"),
    )
)
def test_encode_copy_value(input_value, expected_value):
    assert encode_copy_value(input_value) == expected_value


def test_invalid_escape_sequence():
    with pytest.raises(ValueError):
        decode_copy_value("\\")
    with pytest.raises(ValueError):
        decode_copy_value("\\X")


def test_decode_map_contents():
    assert DECODE_MAP['\\b'] == '\b'
    assert DECODE_MAP['\\n'] == '\n'
    assert DECODE_MAP['\\t'] == '\t'
    assert DECODE_MAP['\\\\'] == '\\'
    assert DECODE_MAP['\\0'] == '\0'
    assert DECODE_MAP['\\74'] == '\74'
    assert DECODE_MAP['\\x0'] == '\0'
    assert DECODE_MAP['\\xa'] == '\x0a'
    assert DECODE_MAP['\\xA'] == '\x0a'
    assert DECODE_MAP['\\x00'] == '\0'
    assert DECODE_MAP['\\xa3'] == '\xa3'
    assert DECODE_MAP['\\xA3'] == '\xa3'
    assert DECODE_MAP['\\xAb'] == '\xab'
    assert DECODE_MAP['\\xaB'] == '\xab'
    assert DECODE_MAP['\\xff'] == '\xff'

    assert '\\' not in DECODE_MAP,  "Unterminated escape is not mapped"
    assert '\\z' not in DECODE_MAP,  "Invalid escape sequences are not mapped"

    assert len(DECODE_MAP) == 1097



import re
import shlex
import subprocess  # nosec B404

import yaml


def get_diff_content():
    """Get differences.

    Check for staged migration files and return diff content,
    if it exists.
    """
    result = subprocess.run(  # nosec B603
        shlex.split("git diff --cached --name-only"),
        capture_output=True,
        text=True,
        check=True,
        shell=False,
    )
    staged_files = result.stdout.strip().split("\n")
    migration_files = [
        f for f in staged_files if "migrations/" in f and f.endswith(".py")
    ]
    if not migration_files:
        return []
    diff_result = subprocess.run(  # nosec B603
        shlex.split("git diff --cached") + migration_files,
        capture_output=True,
        text=True,
        check=True,
        shell=False,
    )
    return diff_result.stdout


def extract_new_fields(diff_content):
    """Extract model and field names.

    Extract model and field names from AddField operations in
    staged migrations.
    """
    add_field_pattern = (
        r"migrations\.AddField\([^)]*"
        r"model_name\s*=\s*['\"](\w+)['\"]"
        r"[^)]*name\s*=\s*['\"](\w+)['\"]"
    )
    return re.findall(add_field_pattern, diff_content, re.DOTALL)


def extract_new_models(diff_content):
    """Extract model names from CreateModel operations."""
    return re.findall(
        r"migrations\.CreateModel\(.*?name\s*=\s*['\"](\w+)['\"]",
        diff_content,
        re.DOTALL,
    )


def check_fields_in_config(new_fields, config_path, writer):
    """Check which fields are missing from anonymiser config."""
    try:
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)
    except FileNotFoundError as e:
        writer.warning(f"Warning: Could not read config file: {e}")
        return []
    strategy = config.get("strategy", {})
    missing = []
    for model_name, field_name in new_fields:
        if model_name not in strategy or field_name not in strategy[model_name]:
            missing.append(f"{model_name}.{field_name}")
    return missing



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import argparse
import codecs
import os
import sys

import six

from .config import Configuration
from .dump import run


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(
        prog=(argv[0] if len(argv) else "database-sanitizer"),
        description="Sanitizes contents of databases.",
    )
    parser.add_argument(
        "--config",
        "-c",
        type=str,
        dest="config",
        help="Path to the sanitizer configuration file.",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        dest="output",
        help=(
            "Path to the file where the sanitized database will be written "
            "into. If omitted, standard output will be used instead."
        ),
    )
    parser.add_argument(
        "url",
        help="Database URL to which to connect into and sanitize contents.",
    )

    args = parser.parse_args(args=argv[1:])
    output = sys.stdout
    if six.PY2:
        output = codecs.getwriter("utf-8")(output)
    config = None

    if args.config:
        conf_dir = os.path.realpath(os.path.dirname(args.config))
        sys.path.insert(0, conf_dir)
        config = Configuration.from_file(args.config)
    if args.output:
        output = open(args.output, "w")

    try:
        run(
            url=args.url,
            output=output,
            config=config,
        )
    finally:
        if args.output:
            output.close()


if __name__ == "__main__":
    main()



from database_sanitizer import session
from database_sanitizer.sanitizers import user


def setup_module():
    session.reset(b'not-so-secret-key')


def test_sanitize_email():
    assert user.sanitize_email(None) is None
    assert user.sanitize_email('') == ''
    assert user.sanitize_email('test@example.com') == (
        'zoe.burke@xce13103b.sanitized.net')
    assert user.sanitize_email('test2@example.com') == (
        'Melanie.Pratt@x4feb7f40.sanitized.net')
    assert user.sanitize_email('test@example.com') == (
        'zoe.burke@xce13103b.sanitized.net')
    assert user.sanitize_email('test3@example.com') == (
        'irene.archer@x3d2e92ec.sanitized.net')
    assert user.sanitize_email(' test3@example.com  ') == (
        'irene.archer@x3d2e92ec.sanitized.net')


def test_sanitize_username():
    assert user.sanitize_username(None) is None
    assert user.sanitize_username('') == ''
    assert user.sanitize_username('John.Doe') == 'billyda979417'
    assert user.sanitize_username('JaneSmith') == 'helena34a7a0b'
    assert user.sanitize_username('john-smith') == 'arthurc5a84ec'
    assert user.sanitize_username('john-smith ') == 'douglas8d3b8d5e'
    assert user.sanitize_username('john smith ') == 'katyfdab90cc'


def test_sanitize_full_name_en_gb():
    assert user.sanitize_full_name_en_gb(None) is None
    assert user.sanitize_full_name_en_gb('') == ''
    assert user.sanitize_full_name_en_gb('John Doe') == 'Francis Walker'
    assert user.sanitize_full_name_en_gb('Jane Smith') == 'Declan Burke'
    assert user.sanitize_full_name_en_gb('John Smith') == 'Lawrence Norton'
    assert user.sanitize_full_name_en_gb('john smith ') == 'Lawrence Norton'


def test_sanitize_given_name_en_gb():
    assert user.sanitize_given_name_en_gb(None) is None
    assert user.sanitize_given_name_en_gb('') == ''
    assert user.sanitize_given_name_en_gb('John') == 'Cheryl'
    assert user.sanitize_given_name_en_gb('Jane') == 'Andrea'
    assert user.sanitize_given_name_en_gb('Foo bar') == 'Elliott'
    assert user.sanitize_given_name_en_gb('  Foo BAR ') == 'Elliott'


def test_sanitize_surname_en_gb():
    assert user.sanitize_surname_en_gb(None) is None
    assert user.sanitize_surname_en_gb('') == ''
    assert user.sanitize_surname_en_gb('Doe') == 'Bibi'
    assert user.sanitize_surname_en_gb('Smith') == 'Duffy'
    assert user.sanitize_surname_en_gb('Anderson') == 'Hodgson'
    assert user.sanitize_surname_en_gb('andersOn ') == 'Hodgson'


def test_sanitize_email_resets_on_session_reset():
    assert user.sanitize_email('test@example.com') == (
        'zoe.burke@xce13103b.sanitized.net')
    session.reset()
    assert user.sanitize_email('test@example.com') != (
        'zoe.burke@xce13103b.sanitized.net')
    session.reset(b'not-so-secret-key')



from unittest.mock import Mock, mock_open

from db_anonymiser.management.commands._check_migration_fields_helpers import (
    check_fields_in_config,
    extract_new_fields,
    extract_new_models,
)


def test_get_diff_content_no_migrations(mocker):
    mock_subprocess = mocker.patch("subprocess.run")
    mock_subprocess.return_value = Mock(stdout="", returncode=0)
    result = extract_new_fields("")
    assert result == []


def test_extract_new_fields_no_addfield_operations():
    result = extract_new_fields("# Migration with no AddField")
    assert result == []


def test_extract_new_fields_single_field():
    diff_content = """
+    operations = [
+        migrations.AddField(
+            model_name='user',
+            name='email',
+            field=models.EmailField(),
+        ),
+    ]
"""
    result = extract_new_fields(diff_content)
    assert result == [("user", "email")]


def test_extract_new_fields_multiple_fields():
    diff_content = """
+        migrations.AddField(
+            model_name='user',
+            name='email',
+            field=models.EmailField(),
+        ),
+        migrations.AddField(
+            model_name='report',
+            name='sensitive_pii',
+            field=models.CharField(max_length=255),
+        ),
"""
    result = extract_new_fields(diff_content)
    assert result == [("user", "email"), ("report", "sensitive_pii")]


def test_extract_new_models_no_create_model():
    result = extract_new_models("# Migration with no CreateModel")
    assert result == []


def test_extract_new_models_single_model():
    diff_content = """
+        migrations.CreateModel(
+            name='Customer',
+            fields=[
+                ('id', models.BigAutoField(auto_created=True, primary_key=True)),
+                ('name', models.CharField(max_length=100)),
+            ],
+        ),
"""
    result = extract_new_models(diff_content)
    assert result == ["Customer"]


def test_extract_new_models_multiple_models():
    diff_content = """
+        migrations.CreateModel(
+            name='Employee',
+            fields=[],
+        ),
+        migrations.CreateModel(
+            name='Department',
+            fields=[],
+        ),
"""
    result = extract_new_models(diff_content)
    assert result == ["Employee", "Department"]


def test_extract_new_models_ignores_add_field():
    diff_content = """
+        migrations.CreateModel(
+            name='Vendor',
+            fields=[],
+        ),
+        migrations.AddField(
+            model_name='report',
+            name='vendor_notes',
+            field=models.TextField(blank=True),
+        ),
"""
    result = extract_new_models(diff_content)
    assert result == ["Vendor"]


def test_check_fields_in_config_all_present(mocker):
    config_content = """
strategy:
  user:
    email: faker.email
  report:
    sensitive_pii: faker.name
"""
    mocker.patch("builtins.open", mock_open(read_data=config_content))
    mock_writer = Mock()
    result = check_fields_in_config(
        [("user", "email"), ("report", "sensitive_pii")], "config.yml", mock_writer
    )
    assert result == []


def test_check_fields_in_config_missing_field(mocker):
    config_content = """
strategy:
  user:
    email: faker.email
"""
    mocker.patch("builtins.open", mock_open(read_data=config_content))
    mock_writer = Mock()
    result = check_fields_in_config(
        [("user", "email"), ("user", "phone")], "config.yml", mock_writer
    )
    assert result == ["user.phone"]


def test_check_fields_in_config_missing_model(mocker):
    config_content = """
strategy:
  user:
    email: faker.email
"""
    mocker.patch("builtins.open", mock_open(read_data=config_content))
    mock_writer = Mock()
    result = check_fields_in_config(
        [("report", "sensitive_pii")], "config.yml", mock_writer
    )
    assert result == ["report.sensitive_pii"]


def test_check_fields_in_config_file_not_found(mocker):
    mocker.patch("builtins.open", side_effect=FileNotFoundError("Config not found"))
    mock_writer = Mock()
    result = check_fields_in_config([("user", "email")], "config.yml", mock_writer)
    assert result == []
    mock_writer.warning.assert_called_once_with(
        "Warning: Could not read config file: Config not found"
    )


def test_check_fields_in_config_empty_strategy(mocker):
    mocker.patch("builtins.open", mock_open(read_data="strategy: {}"))
    mock_writer = Mock()
    result = check_fields_in_config([("user", "email")], "config.yml", mock_writer)
    assert result == ["user.email"]



import datetime

import mock

from database_sanitizer.sanitizers import times


class _FakeDateTime(datetime.datetime):
    @staticmethod
    def now():
        return datetime.datetime(2018, 1, 1, 12, 00, 00)


@mock.patch('random.randint', return_value=42005)
@mock.patch.object(datetime, 'datetime', _FakeDateTime)
def test_sanitize_random_past_timestamp(randint_mock):
    assert times.sanitize_random_past_timestamp('old') == (
        '2018-01-01T11:59:17.995000')



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import random
import string

CHARACTERS = string.ascii_letters + string.digits


def sanitize_empty(value):
    """
    Built-in sanitizer which replaces the original value with empty string.
    """
    return None if value is None else ""


def sanitize_zfill(value):
    """
    Built-in sanitizer which replaces the original value with zeros.
    """
    return None if value is None else "".zfill(len(value))


def sanitize_random(value):
    """
    Random string of same length as the given value.
    """
    if not value:
        return value
    return ''.join(random.choice(CHARACTERS) for _ in range(len(value)))



import pytest

from django.core.management import call_command


HELPERS_PATH = (
    "django_db_anonymiser.db_anonymiser.management.commands"
    "._check_migration_fields_helpers"
)
COMMAND_PATH = (
    "django_db_anonymiser.db_anonymiser.management.commands" ".check_migration_fields"
)


def test_check_migration_fields_no_diff_content(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value=None)
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    call_command("check_migration_fields")
    mock_writer.error.assert_not_called()


def test_check_migration_fields_no_fields_or_models(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="# no ops")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=[])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    call_command("check_migration_fields")
    mock_writer.error.assert_not_called()


def test_check_migration_fields_all_fields_in_config(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[("user", "email")])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.check_fields_in_config", return_value=[])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch("django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "config.yml")
    call_command("check_migration_fields")
    mock_writer.error.assert_not_called()


def test_check_migration_fields_missing_fields_yes(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[("user", "phone")])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.check_fields_in_config", return_value=["user.phone"])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch(
        "django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "/path/to/config.yml"
    )
    mocker.patch("builtins.input", return_value="y")
    call_command("check_migration_fields")
    mock_writer.success.assert_called_once()


def test_check_migration_fields_missing_fields_no(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[("user", "phone")])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.check_fields_in_config", return_value=["user.phone"])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch(
        "django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "/path/to/config.yml"
    )
    mocker.patch("builtins.input", return_value="n")
    with pytest.raises(SystemExit) as exc_info:
        call_command("check_migration_fields")
    assert exc_info.value.code == 1


def test_check_migration_fields_missing_fields_invalid(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[("user", "phone")])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.check_fields_in_config", return_value=["user.phone"])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch(
        "django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "/path/to/config.yml"
    )
    mocker.patch("builtins.input", return_value="maybe")
    with pytest.raises(SystemExit) as exc_info:
        call_command("check_migration_fields")
    assert exc_info.value.code == 1
    assert (
        "\n❌ Invalid response. Commit blocked.\n"
        in mock_writer.error.call_args_list[-1][0][0]
    )


def test_check_migration_fields_new_model_prompts_user(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=["Customer"])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch("django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "config.yml")
    mocker.patch("builtins.input", return_value="y")
    call_command("check_migration_fields")
    warning_calls = [call[0][0] for call in mock_writer.warning.call_args_list]
    assert any("Customer" in call for call in warning_calls)
    mock_writer.success.assert_called_once()


def test_check_migration_fields_new_model_blocked(mocker, mock_writer):
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(f"{HELPERS_PATH}.extract_new_fields", return_value=[])
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=["Customer"])
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch("django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "config.yml")
    mocker.patch("builtins.input", return_value="n")
    with pytest.raises(SystemExit) as exc_info:
        call_command("check_migration_fields")
    assert exc_info.value.code == 1


def test_check_migration_fields_new_model_and_missing_fields(mocker, mock_writer):
    """Both CreateModel and missing AddField fields — single prompt covers both."""
    mocker.patch(f"{HELPERS_PATH}.get_diff_content", return_value="diff content")
    mocker.patch(
        f"{HELPERS_PATH}.extract_new_fields", return_value=[("report", "notes")]
    )
    mocker.patch(f"{HELPERS_PATH}.extract_new_models", return_value=["Customer"])
    mocker.patch(
        f"{HELPERS_PATH}.check_fields_in_config", return_value=["report.notes"]
    )
    mocker.patch(f"{COMMAND_PATH}.helpers.make_writers", return_value=mock_writer)
    mocker.patch("django.conf.settings.DB_ANONYMISER_CONFIG_LOCATION", "config.yml")
    mocker.patch("builtins.input", return_value="y")
    call_command("check_migration_fields")
    # Only one prompt despite two issues
    assert mock_writer.success.call_count == 1



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import importlib

import six
import yaml

__all__ = ("Configuration", "ConfigurationError")

SKIP_ROWS_CONFIG_VALUE = "skip_rows"
MYSQLDUMP_DEFAULT_PARAMETERS = ["--single-transaction"]
PG_DUMP_DEFAULT_PARAMETERS = []


class ConfigurationError(ValueError):
    """
    Custom exception type used to indicate configuration file errors.
    """


class Configuration(object):
    """
    Object representation of database sanitizer configuration, usually read
    from a YAML file.
    """
    def __init__(self):
        self.sanitizers = {}
        self.skip_rows_for_tables = []
        self.addon_packages = []
        self.mysqldump_params = []
        self.pg_dump_params = []

    @classmethod
    def from_file(cls, filename):
        """
        Reads configuration from given path to a file in local file system and
        returns parsed version of it.

        :param filename: Path to the YAML file in local file system where the
                         configuration will be read from.
        :type filename: str

        :return: Configuration instance parsed from given configuration file.
        :rtype: Configuration
        """
        instance = cls()

        with open(filename, "rb") as file_stream:
            config_data = yaml.safe_load(file_stream)

        instance.load(config_data)

        return instance

    def load(self, config_data):
        """
        Loads sanitizers according to rulesets defined in given already parsed
        configuration file.

        :param config_data: Already parsed configuration data, as dictionary.
        :type config_data: dict[str,any]
        """
        if not isinstance(config_data, dict):
            raise ConfigurationError(
                "Configuration data is %s instead of dict." % (
                    type(config_data),
                )
            )

        self.load_addon_packages(config_data)
        self.load_sanitizers(config_data)
        self.load_dump_extra_parameters(config_data)

    def load_dump_extra_parameters(self, config_data):
        """
        Loads extra parameters for mysqldump and/or pg_dump CLI usage. These
        parameters should be added to the mysqldump and/or pg_dump command call
        when taking a dump.

        :param config_data: Already parsed configuration data, as dictionary.
        :type config_data: dict[str,any]
        """
        section_config = config_data.get("config", {})
        if not isinstance(section_config, dict):
            raise ConfigurationError(
                "'config' is %s instead of dict" % (
                    type(section_config),
                ),
            )

        section_extra_parameters = section_config.get("extra_parameters", {})
        if not isinstance(section_extra_parameters, dict):
            raise ConfigurationError(
                "'config.extra_parameters' is %s instead of dict" % (
                    type(section_extra_parameters),
                ),
            )

        mysqldump_params = section_extra_parameters.get("mysqldump", MYSQLDUMP_DEFAULT_PARAMETERS)
        if not isinstance(mysqldump_params, list):
            raise ConfigurationError(
                "'config.extra_parameters.mysqldump' is %s instead of list" % (
                    type(mysqldump_params),
                ),
            )

        pg_dump_params = section_extra_parameters.get("pg_dump", PG_DUMP_DEFAULT_PARAMETERS)
        if not isinstance(pg_dump_params, list):
            raise ConfigurationError(
                "'config.extra_parameters.pg_dump' is %s instead of list" % (
                    type(pg_dump_params),
                ),
            )

        self.mysqldump_params = mysqldump_params
        self.pg_dump_params = pg_dump_params

    def load_addon_packages(self, config_data):
        """
        Loads the module paths from which the configuration will attempt to
        load sanitizers from. These must be stored as a list of strings under
        "config.addons" section of the configuration data.

        :param config_data: Already parsed configuration data, as dictionary.
        :type config_data: dict[str,any]
        """
        section_config = config_data.get("config")
        if not isinstance(section_config, dict):
            if section_config is None:
                return
            raise ConfigurationError(
                "'config' is %s instead of dict" % (
                    type(section_config),
                ),
            )

        section_addons = section_config.get("addons", [])
        if not isinstance(section_addons, list):
            raise ConfigurationError(
                "'config.addons' is %s instead of list" % (
                    type(section_addons),
                ),
            )

        for index, module_path in enumerate(section_addons):
            if not isinstance(module_path, str):
                raise ConfigurationError(
                    "Item %d in 'config.addons' is %s instead of string" % (
                        index,
                        type(module_path),
                    ),
                )

        self.addon_packages = list(section_addons)

    def load_sanitizers(self, config_data):
        """
        Loads sanitizers possibly defined in the configuration under dictionary
        called "strategy", which should contain mapping of database tables with
        column names mapped into sanitizer function names.

        :param config_data: Already parsed configuration data, as dictionary.
        :type config_data: dict[str,any]
        """
        section_strategy = config_data.get("strategy")
        if not isinstance(section_strategy, dict):
            if section_strategy is None:
                return
            if section_strategy != SKIP_ROWS_CONFIG_VALUE:
                raise ConfigurationError(
                    "'strategy' is %s instead of dict" % (
                        type(section_strategy),
                    ),
                )

        for table_name, column_data in six.iteritems(section_strategy):
            if column_data == SKIP_ROWS_CONFIG_VALUE:
                self.skip_rows_for_tables.append(table_name)
                continue

            if not isinstance(column_data, dict):
                if column_data is None:
                    continue
                raise ConfigurationError(
                    "'strategy.%s' is %s instead of dict" % (
                        table_name,
                        type(column_data),
                    ),
                )

            for column_name, sanitizer_name in six.iteritems(column_data):
                if sanitizer_name is None:
                    continue

                if not isinstance(sanitizer_name, str):
                    raise ConfigurationError(
                        "'strategy.%s.%s' is %s instead of string" % (
                            table_name,
                            column_name,
                            type(sanitizer_name),
                        ),
                    )

                sanitizer_callback = self.find_sanitizer(sanitizer_name)
                sanitizer_key = "%s.%s" % (table_name, column_name)
                self.sanitizers[sanitizer_key] = sanitizer_callback

    def find_sanitizer(self, name):
        """
        Searches for a sanitizer function with given name. The name should
        contain two parts separated from each other with a dot, the first
        part being the module name while the second being name of the function
        contained in the module, when it's being prefixed with "sanitize_".

        The lookup process consists from three attempts, which are:

        1. First package to look the module will be top level package called
           "sanitizers".
        2. Module will be looked under the "addon" packages, if they have been
           defined.
        3. Finally the sanitation function will be looked from the builtin
           sanitizers located in "database_sanitizer.sanitizers" package.

        If none of these provide any results, ConfigurationError will be
        thrown.

        :param name: "Full name" of the sanitation function containing name
                     of the module as well as name of the function.
        :type name: str

        :return: First function which can be imported with the given name.
        :rtype: callable
        """
        # Split the sanitizer name into two parts, one containing the Python
        # module name, while second containing portion of the function name
        # we are looking for.
        name_parts = name.split(".")
        if len(name_parts) < 2:
            raise ConfigurationError(
                "Unable to separate module name from function name in '%s'" % (
                    name,
                ),
            )

        module_name_suffix = ".".join(name_parts[:-1])
        function_name = "sanitize_%s" % (name_parts[-1],)

        # Phase 1: Look for custom sanitizer under a top level package called
        # "sanitizers".
        module_name = "sanitizers.%s" % (module_name_suffix,)
        callback = self.find_sanitizer_from_module(
            module_name=module_name,
            function_name=function_name,
        )
        if callback:
            return callback

        # Phase 2: Look for the sanitizer under "addon" packages, if any of
        # such have been defined.
        for addon_package_name in self.addon_packages:
            module_name = "%s.%s" % (
                addon_package_name,
                module_name_suffix,
            )
            callback = self.find_sanitizer_from_module(
                module_name=module_name,
                function_name=function_name,
            )
            if callback:
                return callback

        # Phase 3: Look from builtin sanitizers.
        module_name = "database_sanitizer.sanitizers.%s" % (module_name_suffix,)
        callback = self.find_sanitizer_from_module(
            module_name=module_name,
            function_name=function_name,
        )
        if callback:
            return callback

        # Give up.
        raise ConfigurationError("Unable to find sanitizer called '%s'" % (
            name,
        ))

    @staticmethod
    def find_sanitizer_from_module(module_name, function_name):
        """
        Attempts to find sanitizer function from given module. If the module
        cannot be imported, or function with given name does not exist in it,
        nothing will be returned by this method. Otherwise the found sanitizer
        function will be returned.

        :param module_name: Name of the module to import the function from.
        :type module_name: str

        :param function_name: Name of the function to look for inside the
                              module.
        :type function_name: str

        :return: Sanitizer function found from the module, if it can be
                 imported and it indeed contains function with the given name.
                 Otherwise None will be returned instead.
        :rtype: callback|None
        """
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            return None

        # Look for the function inside the module. At this point it could be
        # pretty much anything.
        callback = getattr(module, function_name, None)

        # Function does not exist in this module? Give up.
        if callback is None:
            return None

        # It's actually callable function? Return it.
        if callable(callback):
            return callback

        # Sanitizer seems to be something else than a function. Throw an
        # exception to report such problem.
        raise ConfigurationError("'%s' in '%s' is %s instead of function" % (
            function_name,
            module_name,
            type(callback),
        ))

    def get_sanitizer_for(self, table_name, column_name):
        """
        Get sanitizer for given table and column name.

        :param table_name: Name of the database table.
        :type table_name: str

        :param column_name: Name of the database column.
        :type column_name: str

        :return: Sanitizer function or None if nothing is configured
        :rtype: Optional[Callable[[Optional[str]], Optional[str]]]
        """
        sanitizer_key = "%s.%s" % (table_name, column_name)
        return self.sanitizers.get(sanitizer_key)

    def sanitize(self, table_name, column_name, value):
        """
        Sanitizes given value extracted from the database according to the
        sanitation configuration.

        TODO: Add support for dates, booleans and other types found in SQL than
        string.

        :param table_name: Name of the database table from which the value is
                           from.
        :type table_name: str

        :param column_name: Name of the database column from which the value is
                            from.
        :type column_name: str

        :param value: Value from the database, either in text form or None if
                      the value is null.
        :type value: str|None

        :return: Sanitized version of the given value.
        :rtype: str|None
        """
        sanitizer_callback = self.get_sanitizer_for(table_name, column_name)
        return sanitizer_callback(value) if sanitizer_callback else value



# -*- coding: utf-8 -*-

from collections import namedtuple

import mock
import pytest

from .. import config
from ..config import Configuration, ConfigurationError


@mock.patch.object(config, 'open')
@mock.patch('yaml.safe_load')
def test_from_file(mocked_yaml_load, mocked_open):
    mocked_yaml_load.return_value = {}

    Configuration.from_file('filename.yml')

    assert mocked_open.call_args == (('filename.yml', 'rb'), {})
    opened_file = mocked_open.return_value.__enter__.return_value
    assert mocked_yaml_load.call_args == ((opened_file,), {})


def test_load_config_data_must_be_dict():
    config = Configuration()
    config.load({})
    with pytest.raises(ConfigurationError):
        config.load(config_data="test")


def test_load_dump_extra_parameters():
    config = Configuration()

    config.load_dump_extra_parameters({})
    assert config.mysqldump_params == ["--single-transaction"]
    assert config.pg_dump_params == []

    with pytest.raises(ConfigurationError):
        config.load_dump_extra_parameters({"config": "test"})

    config.load_dump_extra_parameters({"config": {}})
    assert config.mysqldump_params == ["--single-transaction"]
    assert config.pg_dump_params == []

    with pytest.raises(ConfigurationError):
        config.load_dump_extra_parameters({"config": {
            "extra_parameters": "test"
        }})

    with pytest.raises(ConfigurationError):
        config.load_dump_extra_parameters({"config": {
            "extra_parameters": [True]
        }})

    with pytest.raises(ConfigurationError):
        config.load_dump_extra_parameters({"config": {
            "extra_parameters": {
                "mysqldump": "hernekeitto",
            },
        }})

    with pytest.raises(ConfigurationError):
        config.load_dump_extra_parameters({"config": {
            "extra_parameters": {
                "pg_dump": "viina",
            },
        }})

    config.load_dump_extra_parameters({"config": {
        "extra_parameters": {
            "mysqldump": ["--double-transaction"],
            "pg_dump": ["--exclude-table=something"],
        },
    }})
    assert config.mysqldump_params == ["--double-transaction"]
    assert config.pg_dump_params == ["--exclude-table=something"]


def test_load_addon_packages():
    config = Configuration()

    config.load_addon_packages({})
    assert config.addon_packages == []

    with pytest.raises(ConfigurationError):
        config.load_addon_packages({"config": "test"})

    config.load_addon_packages({"config": {}})
    assert config.addon_packages == []

    with pytest.raises(ConfigurationError):
        config.load_addon_packages({"config": {"addons": "test"}})

    with pytest.raises(ConfigurationError):
        config.load_addon_packages({"config": {"addons": [True]}})

    config.load_addon_packages({"config": {
        "addons": [
            "test1",
            "test2",
            "test3",
        ],
    }})
    assert config.addon_packages == ["test1", "test2", "test3"]


def test_load_sanitizers():
    config = Configuration()

    with pytest.raises(ConfigurationError):
        config.load_sanitizers({"strategy": "test"})

    with pytest.raises(ConfigurationError):
        config.load_sanitizers({"strategy": {"test": "test"}})

    def mock_find_sanitizer(*args):
        return lambda value: value

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer",
                    side_effect=mock_find_sanitizer):
        with pytest.raises(ConfigurationError):
            config.load_sanitizers({"strategy": {"table1": {"column1": True}}})

        config.load_sanitizers({"strategy": {
            "table1": {
                "column1": None,
                "column2": "test.test",
            },
            "table2": {
                "column1": "test.test",
            },
            "table3": None,
        }})

    assert "table1.column1" not in config.sanitizers
    assert "table1.column2" in config.sanitizers
    assert "table2.column1" in config.sanitizers


def test_table_skip_rows_configuration():
    config = Configuration()

    with pytest.raises(ConfigurationError):
        config.load_sanitizers({"strategy": "test"})

    def mock_find_sanitizer(*args):
        return lambda value: value

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer",
                    side_effect=mock_find_sanitizer):

        config.load_sanitizers({"strategy": {
            "table1": "skip_rows",
            "table2": {
                "column1": "test",
            }
        }})

    assert "table2.column1" in config.sanitizers
    assert "table1" in config.skip_rows_for_tables


def test_find_sanitizer():
    config = Configuration()

    with pytest.raises(ConfigurationError):
        config.find_sanitizer("test")

    def mock_find_sanitizer_from_module1(module_name, function_name):
        assert module_name == "sanitizers.test"
        assert function_name == "sanitize_test"
        return lambda value: value

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer_from_module",
                    side_effect=mock_find_sanitizer_from_module1):
        assert config.find_sanitizer("test.test") is not None

    def mock_find_sanitizer_from_module2(module_name, function_name):
        assert module_name in ("sanitizers.test", "addon.test")
        assert function_name == "sanitize_test"
        if module_name.startswith("addon."):
            return lambda value: value
        else:
            return None

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer_from_module",
                    side_effect=mock_find_sanitizer_from_module2):
        config.addon_packages = ("addon",)
        assert config.find_sanitizer("test.test") is not None

    def mock_find_sanitizer_from_module3(module_name, function_name):
        assert module_name in (
            "sanitizers.test",
            "addon.test",
            "database_sanitizer.sanitizers.test",
        )
        assert function_name == "sanitize_test"
        if module_name.startswith("database_sanitizer."):
            return lambda value: value
        else:
            return None

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer_from_module",
                    side_effect=mock_find_sanitizer_from_module3):
        assert config.find_sanitizer("test.test") is not None

    def mock_find_sanitizer_from_module4(module_name, function_name):
        return None

    with mock.patch("database_sanitizer.config.Configuration.find_sanitizer_from_module",
                    side_effect=mock_find_sanitizer_from_module4):
        with pytest.raises(ConfigurationError):
            config.find_sanitizer("test.test")


def test_find_sanitizer_from_module():
    def mock_import1(module_name):
        assert module_name == "test"
        raise ImportError("Should be catched")

    with mock.patch("importlib.import_module", side_effect=mock_import1):
        assert Configuration.find_sanitizer_from_module("test", "test") is None

    mock_module_type = namedtuple("mock_module", ("test",))

    def mock_import2(module_name):
        assert module_name == "test"
        return mock_module_type(test=None)

    with mock.patch("importlib.import_module", side_effect=mock_import2):
        assert Configuration.find_sanitizer_from_module("test", "test") is None

    def mock_import3(module_name):
        assert module_name == "test"
        return mock_module_type(test=lambda value: value)

    with mock.patch("importlib.import_module", side_effect=mock_import3):
        assert Configuration.find_sanitizer_from_module("test", "test") is not None

    def mock_import4(module_name):
        assert module_name == "test"
        return mock_module_type(test="test")

    with mock.patch("importlib.import_module", side_effect=mock_import4):
        with pytest.raises(ConfigurationError):
            Configuration.find_sanitizer_from_module("test", "test")


def test_sanitize():
    config = Configuration()
    config.sanitizers["a.a"] = lambda value: value.upper()
    config.sanitizers["a.b"] = lambda value: value[::-1]

    assert config.sanitize("a", "a", "test") == "TEST"
    assert config.sanitize("a", "b", "test") == "tset"
    assert config.sanitize("a", "c", "test") == "test"



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import importlib

from six.moves.urllib import parse as urlparse

from .. import session

SUPPORTED_DATABASE_MODULES = {
    "mysql": "django_db_anonymiser.database_sanitizer.dump.mysql",
    "postgres": "django_db_anonymiser.database_sanitizer.dump.postgres",
    "postgresql": "django_db_anonymiser.database_sanitizer.dump.postgres",
    "postgis": "django_db_anonymiser.database_sanitizer.dump.postgres",
}


# Register supported database schemes.
for scheme in SUPPORTED_DATABASE_MODULES.keys():
    urlparse.uses_netloc.append(scheme)


def run(url, output, config):
    """
    Extracts database dump from given database URL and outputs sanitized
    copy of it into given stream.

    :param url: URL to the database which is to be sanitized.
    :type url: str

    :param output: Stream where sanitized copy of the database dump will be
                   written into.
    :type output: file

    :param config: Optional sanitizer configuration to be used for sanitation
                   of the values stored in the database.
    :type config: database_sanitizer.config.Configuration|None
    """
    parsed_url = urlparse.urlparse(url)
    db_module_path = SUPPORTED_DATABASE_MODULES.get(parsed_url.scheme)
    if not db_module_path:
        raise ValueError("Unsupported database scheme: '%s'" % (parsed_url.scheme,))
    db_module = importlib.import_module(db_module_path)
    session.reset()
    for line in db_module.sanitize(url=parsed_url, config=config):
        output.write(line + "\n")



from __future__ import unicode_literals

from six import text_type

from database_sanitizer.session import hash_text_to_int, hash_text_to_ints


def sanitize_email(value):
    if not value:
        return value
    (num1, num2, num3) = hash_text_to_ints(value.strip(), [16, 16, 32])
    given_name = given_names[num1 % given_names_count]
    surname = surnames[num2 % surnames_count]
    case_convert = (text_type.lower if num3 % 8 > 0 else lambda x: x)
    return '{first}.{last}@x{num:x}.sanitized.net'.format(
        first=case_convert(given_name),
        last=case_convert(surname).replace("'", ''),
        num=num3)


def sanitize_username(value):
    if not value:
        return value
    (num1, num2) = hash_text_to_ints(value, [16, 32])
    return '{}{:x}'.format(given_names[num1 % given_names_count].lower(), num2)


def sanitize_full_name_en_gb(value):
    if not value:
        return value
    (num1, num2) = hash_text_to_ints(value.strip().lower(), [16, 16])
    return '{} {}'.format(
        given_names[num1 % given_names_count], surnames[num2 % surnames_count])


def sanitize_given_name_en_gb(value):
    if not value:
        return value
    num = hash_text_to_int(value.strip().lower())
    return given_names[num % given_names_count]


def sanitize_surname_en_gb(value):
    if not value:
        return value
    num = hash_text_to_int(value.strip().lower())
    return surnames[num % surnames_count]


given_names = """
Aaron Abbie Abdul Abigail Adam Adrian Aimee Alan Albert Alex
Alexander Alexandra Alice Alison Allan Amanda Amber Amelia Amy Andrea
Andrew Angela Ann Anna Anne Annette Anthony Antony Arthur Ashleigh
Ashley Barbara Barry Ben Benjamin Bernard Beth Bethan Bethany Beverley
Billy Bradley Brandon Brenda Brett Brian Bruce Bryan Callum Cameron Carl
Carly Carol Carole Caroline Carolyn Catherine Charlene Charles Charlie
Charlotte Chelsea Cheryl Chloe Christian Christine Christopher Claire
Clare Clifford Clive Colin Connor Conor Craig Dale Damian Damien Daniel
Danielle Danny Darren David Dawn Dean Deborah Debra Declan Denis Denise
Dennis Derek Diana Diane Dominic Donald Donna Dorothy Douglas Duncan
Dylan Edward Eileen Elaine Eleanor Elizabeth Ellie Elliot Elliott Emily
Emma Eric Fiona Frances Francesca Francis Frank Frederick Gail Gareth
Garry Gary Gavin Gemma Geoffrey George Georgia Georgina Gerald Geraldine
Gerard Gillian Glen Glenn Gordon Grace Graeme Graham Gregory Guy Hannah
Harriet Harry Hayley Hazel Heather Helen Henry Hilary Hollie Holly
Howard Hugh Iain Ian Irene Jack Jacob Jacqueline Jade Jake James Jamie
Jane Janet Janice Jasmine Jason Jay Jayne Jean Jeffrey Jemma Jenna
Jennifer Jeremy Jessica Jill Joan Joanna Joanne Jodie Joe Joel John
Jonathan Jordan Joseph Josephine Josh Joshua Joyce Judith Julia Julian
Julie June Justin Karen Karl Kate Katherine Kathleen Kathryn Katie Katy
Kayleigh Keith Kelly Kenneth Kerry Kevin Kieran Kim Kimberley Kirsty
Kyle Laura Lauren Lawrence Leah Leanne Lee Leigh Leon Leonard Lesley
Leslie Lewis Liam Linda Lindsey Lisa Lorraine Louis Louise Lucy Luke
Lydia Lynda Lynn Lynne Malcolm Mandy Marc Marcus Margaret Maria Marian
Marie Marilyn Marion Mark Martin Martyn Mary Mathew Matthew Maureen
Maurice Max Megan Melanie Melissa Michael Michelle Mitchell Mohamed
Mohammad Mohammed Molly Naomi Natalie Natasha Nathan Neil Nicholas
Nicola Nicole Nigel Norman Oliver Olivia Owen Paige Pamela Patricia
Patrick Paul Paula Pauline Peter Philip Phillip Rachael Rachel Raymond
Rebecca Reece Rhys Richard Ricky Rita Robert Robin Roger Ronald Rosemary
Rosie Ross Roy Russell Ruth Ryan Sally Sam Samantha Samuel Sandra Sara
Sarah Scott Sean Shane Shannon Sharon Shaun Sheila Shirley Sian Simon
Sophie Stacey Stanley Stephanie Stephen Steven Stewart Stuart Susan
Suzanne Sylvia Terence Teresa Terry Thomas Timothy Tina Toby Tom Tony
Tracey Tracy Trevor Valerie Vanessa Victor Victoria Vincent Wayne Wendy
William Yvonne Zoe
""".strip().split()


surnames = """
Abbott Adams Ahmed Akhtar Alexander Ali Allan Allen Anderson Andrews
Archer Armstrong Arnold Ashton Atkins Atkinson Austin Bailey Baker
Baldwin Ball Banks Barber Barker Barlow Barnes Barnett Barrett Barry
Bartlett Barton Bates Baxter Begum Bell Bennett Benson Bentley Berry
Bevan Bibi Birch Bird Bishop Black Blackburn Bolton Bond Booth Bowen
Boyle Bradley Bradshaw Brady Bray Brennan Briggs Brookes Brooks Brown
Browne Bruce Bryan Bryant Bull Burgess Burke Burns Burrows Burton
Butcher Butler Byrne Cameron Campbell Carey Carpenter Carr Carroll
Carter Cartwright Chadwick Chambers Chan Chandler Chapman Charlton Clark
Clarke Clayton Clements Coates Cole Coleman Coles Collier Collins
Connolly Connor Conway Cook Cooke Cooper Cox Craig Crawford Cross
Cunningham Curtis Dale Daly Daniels Davey Davidson Davies Davis Davison
Dawson Day Dean Dennis Dickinson Dixon Dobson Dodd Doherty Donnelly
Douglas Doyle Duffy Duncan Dunn Dyer Edwards Elliott Ellis Evans Farmer
Farrell Faulkner Ferguson Field Finch Fisher Fitzgerald Fleming Fletcher
Flynn Ford Forster Foster Fowler Fox Francis Franklin Fraser Freeman
French Frost Fry Fuller Gallagher Gardiner Gardner Garner George Gibbons
Gibbs Gibson Gilbert Giles Gill Glover Goddard Godfrey Goodwin Gordon
Gough Gould Graham Grant Gray Green Greenwood Gregory Griffin Griffiths
Hale Hall Hamilton Hammond Hancock Hanson Harding Hardy Hargreaves
Harper Harris Harrison Hart Hartley Harvey Hawkins Hayes Haynes Hayward
Heath Henderson Henry Herbert Hewitt Hicks Higgins Hill Hilton Hodgson
Holden Holland Holloway Holmes Holt Hooper Hope Hopkins Horton Houghton
Howard Howarth Howe Howell Howells Hudson Hughes Humphreys Humphries
Hunt Hunter Hurst Hussain Hutchinson Hyde Ingram Iqbal Jackson James
Jarvis Jenkins Jennings John Johnson Johnston Jones Jordan Joyce Kaur
Kay Kelly Kemp Kennedy Kent Kerr Khan King Kirby Kirk Knight Knowles
Lamb Lambert Lane Law Lawrence Lawson Leach Lee Lees Leonard Lewis
Little Lloyd Long Lord Lowe Lucas Lynch Lyons Macdonald Mahmood Mann
Manning Marsden Marsh Marshall Martin Mason Matthews May McCarthy
McDonald McKenzie McLean Mellor Metcalfe Miah Middleton Miles Miller
Mills Mistry Mitchell Moore Moran Morgan Morley Morris Morrison Morton
Moss Murphy Murray Myers Nash Naylor Nelson Newman Newton Nicholls
Nicholson Nixon Noble Nolan Norman Norris North Norton O'Blake O'Buckley
O'Chamberlain O'Hobbs O'Thompson Oliver Osborne Owen Owens Page Palmer
Parker Parkes Parkin Parkinson Parry Parsons Patel Patterson Payne
Peacock Pearce Pearson Perkins Perry Peters Phillips Pickering Pollard
Poole Pope Porter Potter Potts Powell Power Pratt Preston Price
Pritchard Pugh Quinn Rahman Randall Read Reed Rees Reeves Reid Reynolds
Rhodes Rice Richards Richardson Riley Roberts Robertson Robinson Robson
Rogers Rose Ross Rowe Rowley Russell Ryan Sanders Sanderson Saunders
Savage Schofield Scott Shah Sharp Sharpe Shaw Shepherd Sheppard Short
Simmons Simpson Sims Sinclair Singh Skinner Slater Smart Smith Spencer
Stanley Steele Stephens Stephenson Stevens Stevenson Stewart Stokes
Stone Storey Sullivan Summers Sutton Swift Sykes Talbot Taylor Thomas
Thomson Thornton Thorpe Todd Tomlinson Townsend Tucker Turnbull Turner
Tyler Vaughan Vincent Wade Walker Wall Wallace Wallis Walsh Walters
Walton Ward Warner Warren Waters Watkins Watson Watts Webb Webster Welch
Wells West Weston Wheeler White Whitehead Whitehouse Whittaker Wilkins
Wilkinson Williams Williamson Willis Wilson Winter Wong Wood Woods
Woodward Wright Wyatt Yates Young
""".strip().split()

given_names_count = len(given_names)
surnames_count = len(surnames)
